// Verilog source for SN76489 VGM file player on Altera/Terasic DE-1 board
// 2013 furrtek - furrtek.org

module SN76489_Player (
	SW, 			// Switch inputs
	KEY, 			// Key inputs
	LEDR,			// Red LEDs
	LEDG,			// Green LEDs
	CLOCK_24,	// Clock signals
	CLOCK_27,
	CLOCK_50,
	HEX0, 		// 7-Seg digits
	HEX1, 
	HEX2, 
	HEX3, 
	FL_CE_N, 	// Flash /CE
	FL_OE_N, 	// Flash /OE
	FL_DQ, 		// Flash data
	FL_ADDR, 	// Flash address
	FL_RST_N,	// Flash /RESET
	I2C_SDAT,	//	I2C Data
	I2C_SCLK,	//	I2C Clock
	AUD_DACLRCK,//	Audio CODEC DAC LR Clock
	AUD_DACDAT,	//	Audio CODEC DAC Data
	AUD_BCLK,	//	Audio CODEC Bit-Stream Clock
	AUD_XCK		//	Audio CODEC Chip Clock
	);

	input 	[9:0] SW;
	input		[3:0]	KEY;
	output 	reg [9:0] LEDR;
	output 	reg [7:0] LEDG;
	
	input		[1:0]	CLOCK_24;
	input		[1:0]	CLOCK_27;
	input 	CLOCK_50

	output	[6:0]	HEX0;
	output	[6:0]	HEX1;
	output	[6:0]	HEX2;
	output	[6:0]	HEX3;
	
	output	FL_OE_N;
	output	FL_CE_N;
	output	FL_RST_N;
	input		[7:0] FL_DQ;
	output 	reg [21:0] FL_ADDR;
	
	inout		I2C_SDAT;
	output	I2C_SCLK;
	output	AUD_DACLRCK;
	output	AUD_DACDAT;
	inout		AUD_BCLK;
	output	AUD_XCK;

	reg		[7:0] VGMCMD;
	
	reg		[2:0] GOTCH;
	
	reg		[2:0] READSTEP;

	reg		[15:0] WAIT;
	
	reg		[10:0] slowclk;
	
	reg		[9:0] TONE1_SET;
	reg		[9:0] TONE2_SET;
	reg		[9:0] TONE3_SET;
	reg		[2:0] NOISE_SET;
	
	reg		[3:0] VOL1;
	reg		[3:0] VOL2;
	reg		[3:0] VOL3;
	reg		[3:0] VOL4;
	
	wire		VGA_CTRL_CLK;
	wire		AUD_CTRL_CLK;

	assign 	FL_RST_N = 1;
	assign 	FL_OE_N = 0;
	assign 	FL_CE_N = 0;
	
	assign	AUD_ADCLRCK	= AUD_DACLRCK;
	assign	AUD_XCK = AUD_CTRL_CLK;
	
	always@(negedge KEY[1])
	begin
		FL_ADDR[17]	= ~FL_ADDR[17];
	end
	
	always@(posedge slowclk[8])
	begin
		if (FL_ADDR[16:0] == 0)
		begin
			FL_ADDR[16:0] = 16'h40;
			READSTEP = 0;
			WAIT = 0;
		end
		else if (WAIT == 0)
		begin
			if (READSTEP == 0)
			begin
				if (KEY[1] == 0) FL_ADDR[16:0] = 16'h40;
				VGMCMD[7:0] = FL_DQ[7:0];
				case(VGMCMD[7:0])
					8'h50 : READSTEP <= 1;		// Write to PSG port 
					8'h61 : READSTEP <= 1;		// Wait nn nn samples
					8'h62 :
					begin
						WAIT = 735;
						READSTEP <= 0;		// Wait 735 samples
					end
				endcase
			end
			else if (READSTEP == 1)
			begin
				case(VGMCMD)
					8'h50 :
					begin
						READSTEP <= 0;	// Write to PSG port 
						if (FL_DQ[7] == 1)	// Reg selection/data
						begin
							GOTCH = FL_DQ[6:4];
							case(GOTCH)
								3'b000 :
								begin
									TONE1_SET[3:0] <= FL_DQ[3:0];
									LEDG[0] <= ~LEDG[0];
								end
								3'b001 : VOL1 <= FL_DQ[3:0];
								3'b010 :
								begin
									TONE2_SET[3:0] <= FL_DQ[3:0];
									LEDG[1] <= ~LEDG[1];
								end
								3'b011 : VOL2 <= FL_DQ[3:0];
								3'b100 :
								begin
									TONE3_SET[3:0] <= FL_DQ[3:0];
									LEDG[2] <= ~LEDG[2];
								end
								3'b101 : VOL3 <= FL_DQ[3:0];
								3'b110 :
								begin
									NOISE_SET <= FL_DQ[2:0];
									LEDR[6:5] = NOISE_SET[1:0];
									LEDG[3] <= ~LEDG[3];
								end
								3'b111 :
								begin
									VOL4 <= FL_DQ[3:0];
									LEDG[3] <= ~LEDG[3];
								end
							endcase
						end
						else
						begin									// Data write
							case(GOTCH)
								3'b000 : TONE1_SET[9:4] <= FL_DQ[5:0];
								3'b001 : VOL1 <= FL_DQ[3:0];
								3'b010 : TONE2_SET[9:4] <= FL_DQ[5:0];
								3'b011 : VOL2 <= FL_DQ[3:0];
								3'b100 : TONE3_SET[9:4] <= FL_DQ[5:0];
								3'b101 : VOL3 <= FL_DQ[3:0];
								3'b110 : NOISE_SET <= FL_DQ[2:0];
								3'b111 : VOL4 <= FL_DQ[3:0];
							endcase
						end
					end
					8'h61 :
					begin
						WAIT[7:0] = FL_DQ[7:0];
						READSTEP <= 2;	// Wait nn nn samples
					end
				endcase
			end
			else if (READSTEP == 2)
			begin
				// Wait nn nn samples
				WAIT[15:8] = FL_DQ[7:0];
				READSTEP <= 0;
			end
			if (VGMCMD != 8'h66)
				FL_ADDR[16:0] <= FL_ADDR + 1;
			else
				FL_ADDR[16:0] = 16'h40;
		end
		if (WAIT != 0) WAIT <= WAIT - 1;
	end
	
	always@(posedge CLOCK_24[0])
	begin
		LEDR[9:8] = READSTEP;
		slowclk <= slowclk + 1'b1;
	end

	SEG7_LUT_4 			u0	(HEX0,HEX1,HEX2,HEX3,{VGMCMD[7:0],FL_ADDR[15:8]});

	VGA_Audio_PLL 		u3	(.inclk0(CLOCK_27[0]),.c0(VGA_CTRL_CLK),.c1(AUD_CTRL_CLK));

	I2C_AV_Config 		u7	(
							.iCLK(CLOCK_50),
							.iRST_N(KEY[0]),
							.I2C_SCLK(I2C_SCLK),
							.I2C_SDAT(I2C_SDAT)	);

	AUDIO_DAC 			u8	(
							.oAUD_BCK(AUD_BCLK),
							.oAUD_DATA(AUD_DACDAT),
							.oAUD_LRCK(AUD_DACLRCK),
							.iSrc_Select(SW[9]),
				         .iCLK_18_4(CLOCK_24[0]),
							.iRST_N(KEY[0]),
							.TONE1_SET(TONE1_SET),
							.TONE2_SET(TONE2_SET),
							.TONE3_SET(TONE3_SET),
							.NOISE_SET(NOISE_SET),
							.VOL1(VOL1),
							.VOL2(VOL2),
							.VOL3(VOL3),
							.VOL4(VOL4));
endmodule